@@ -1,35 +1,64 @@
 # -*- coding: utf-8 -*-
-
 import os
+import threading
+import datetime
+import time
+from functools import lru_cache, wraps
 
 import math
 import numpy
 import pandas
 
-from functools import lru_cache
-from .common import *
-
-from dateutil.parser import parse
+from .utils import *
+from .technical_analysis import *
 
 
-class QuoteData:
-    _data_path = None
-    _options = None
+def assert_quote_init(func):
+    @wraps(func)
+    def _wrapper(*args, **kwargs):
+        data_path = QuoteData().get_data_path()
+        if not data_path or not os.path.exists(data_path):
+            print("QuoteData is not initialized.")
+        else:
+            return func(*args, **kwargs)
 
-    def __init__(self, data_path):
+    return _wrapper
+
+
+class QuoteData(object):
+    '''
+    _instance_lock = threading.Lock()
+
+    def __new__(cls, *args, **kwargs):
+        """单例模式"""
+        if not hasattr(cls, "_instance"):
+            with cls._instance_lock:
+                if not hasattr(cls, "_instance"):
+                    cls._instance = object.__new__(cls)
+        return cls._instance
+    '''
+
+    _threading_local = threading.local()
+
+    def __new__(cls, *args, **kwargs):
+        """单例模式-线程隔离"""
+        if not hasattr(cls._threading_local, "_instance"):
+            cls._threading_local._instance = object.__new__(cls)
+        return cls._threading_local._instance
+
+    def __init__(self, data_path=None):
         """本地行情数据类
 
         参数:
-            data_path: 数据路径
+            data_path: 数据目录
 
         """
-        if not data_path or not os.path.exists(data_path):
-            raise Exception('the DATA_PATH is not exists.')
-        self._data_path = data_path
-        self._options = {
-            # 'cache_max_num': 16
-        }
+        if not hasattr(self, '_data_path'):
+            self._data_path = ''
+        if data_path:
+            self.init(data_path)
 
+    @assert_quote_init
     @lru_cache(maxsize=None, typed=True)
     def _fetch_data(self, filename, date_columns=None):
         """获取数据
@@ -66,35 +95,134 @@ class QuoteData:
                 result = f.readlines()
         return result
 
-    def get_option(self, name):
-        """获取可选项
+    def init(self, data_path):
+        """初始化
 
         参数:
-            name: 可选项名
-
-        返回值: 可选项值
+            data_path: 数据目录
 
         """
+        if not data_path or not os.path.exists(data_path):
+            raise Exception('initialization failed. data_path %s is not exists.' % data_path)
+        self._data_path = data_path
 
-        if name in self._options:
-            return self._options[name]
-        else:
-            raise Exception('option %s is invalid.')
+    def get_data_path(self):
+        """获取数据目录
 
-    def set_option(self, name, value):
-        """设置可选项
+        返回值: 数据目录
+
+        """
+        return self._data_path
+
+    @classmethod
+    def _get_price_precision(cls, security) -> int:
+        """获取价格精度
 
         参数:
-            name: 可选项名
-            value: 可选项值
+            security: 字符串或列表
+
+        返回值: 精度或精度列表
 
         """
+        if type(security) is str:
+            n = 2
+            p = security.find('.')
+            if p >= 0:
+                code = security[:p]
+                exch = security[p + 1:]
+                if exch == 'SZ':
+                    if code[0] == '1':
+                        n = 3
+                    elif code[0:3] == '399' or code[0:4] == '9000':
+                        n = 4
+                elif exch == 'SH':
+                    if code[0] == '5':
+                        n = 3
+                    elif code[0:3] == '000' or code[0:4] == '1000':
+                        n = 4
+            return n
+        elif type(security) is list:
+            r = []
+            for s in security:
+                r.append(cls._get_price_precision(s))
+            return r
 
-        if name in self._options:
-            self._options[name] = value
+    @classmethod
+    def _parse_period_str(cls, period) -> (str, int):
+        """分析周期字串
+
+        参数:
+            period: 周期
+
+        返回值: 类型，间隔
+
+        """
+        if period in ['day', 'daily', '1d', 'week', '1w', 'mon', 'month', '1M', 'season', 'year']:
+            period_type = 'day'
+            period_step = 0
+        elif period[-1] == 'm' and period[:-1].isdecimal():
+            period_type = 'mn1'
+            period_step = int(period[:-1])
+        elif period[0:2] == 'mn' and period[2:].isdecimal():
+            period_type = 'mn1'
+            period_step = int(period[2:])
         else:
-            raise Exception('option %s is invalid.')
+            raise Exception('period is invalid.')
+        return period_type, period_step
 
+    def _parse_period_start(self, period, date, count):
+        """分析周期开始时间
+
+        参数:
+            period: 周期
+            date: 参考时间
+            count: 数量
+
+        返回值:
+            参考时间 - count 个周期后的开始时间
+
+        """
+        year_minimum = 1990
+        if period in ['day', 'daily', '1d']:
+            trade_list = self.get_trade_days(end_date=date, count=count)
+            return trade_list[0].timetuple()
+        elif period in ['week', '1w']:
+            trade_list = self.get_trade_days(end_date=date, count=count * 5)
+            week_maps = [x.isocalendar()[0] * 100 + x.isocalendar()[1] for x in trade_list]
+            week_list = list(set(week_maps))
+            week_list.sort()
+            w = week_list[-count]
+            return trade_list[week_maps.index(w)].timetuple()
+        elif period in ['mon', 'month', '1M']:
+            y = date.tm_year
+            m = date.tm_mon + 1 - count
+            while m <= 0:
+                y -= 1
+                m += 12
+            y = max(y, year_minimum)
+            start = datetime.datetime(y, m, 1).timetuple()
+        elif period == 'season':
+            y = date.tm_year
+            s = (date.tm_mon - 1) // 3 + 1 - count
+            while s < 0:
+                y -= 1
+                s += 4
+            y = max(y, year_minimum)
+            start = datetime.datetime(y, s * 3 + 1, 1).timetuple()
+        elif period == 'year':
+            y = date.tm_year + 1 - count
+            y = max(y, year_minimum)
+            start = datetime.datetime(y, 1, 1).timetuple()
+        else:
+            t, n = self._parse_period_str(period)
+            if t != 'mn1':
+                return
+            days = math.ceil(count * n / 240)
+            trade_list = self.get_trade_days(end_date=date, count=days)
+            start = trade_list[0].timetuple()
+        return start
+
+    @assert_quote_init
     def execute_rehabilitation(self, data, fq='pre', fq_vol=False, columns=None) -> pandas.DataFrame:
         """执行复权操作
 
@@ -107,7 +235,7 @@ class QuoteData:
 
         参数:
             data: 输入数据
-            fq: 复权方式或日期
+            fq: 复权方式或复权参考日
             fq_vol: 是否复权成交量：默认值 False
             columns: 需要复权的字段列表
 
@@ -150,7 +278,7 @@ class QuoteData:
                 row['transfer_ratio'] if not math.isnan(row['transfer_ratio']) else 0))
             v = (row['bonus_ratio'] if not math.isnan(row['bonus_ratio']) else 0) / 10
             if fq_cols_px:
-                n = self.get_price_precision(row['code'])
+                n = self._get_price_precision(row['code'])
                 if ts_date <= ts_ref:
                     c = df.loc[(df['code'] == row['code']) & (df['datetime'] < ts_date), fq_cols_px]
                     if len(c) > 0:
@@ -174,38 +302,7 @@ class QuoteData:
                         df.loc[(df['code'] == row['code']) & (df['datetime'] >= ts_date), fq_cols_vm] = c
         return df
 
-    def get_price_precision(self, security) -> int:
-        """获取价格精度
-
-        参数:
-            security: 字符串或列表
-
-        返回值: 精度或精度列表
-
-        """
-        if type(security) is str:
-            n = 2
-            p = security.find('.')
-            if p >= 0:
-                code = security[:p]
-                exch = security[p + 1:]
-                if exch == 'SZ':
-                    if code[0] == '1':
-                        n = 3
-                    elif code[0:3] == '399' or code[0:4] == '9000':
-                        n = 4
-                elif exch == 'SH':
-                    if code[0] == '5':
-                        n = 3
-                    elif code[0:3] == '000' or code[0:4] == '1000':
-                        n = 4
-            return n
-        elif type(security) is list:
-            r = []
-            for s in security:
-                r.append(self.get_price_precision(s))
-            return r
-
+    @assert_quote_init
     def get_quote_date(self) -> time.struct_time:
         """获取行情日期
 
@@ -231,7 +328,7 @@ class QuoteData:
             if not k or not v:
                 continue
             result[k] = v
-
+        # print(result)
         stime = result.get('last_date')
         if not stime:
             raise Exception('the quote_data not found.')
@@ -241,6 +338,7 @@ class QuoteData:
             raise Exception('trade-days is empty.')
         return formatter_st(tdays[0])
 
+    @assert_quote_init
     def get_trade_days(self, start_date=None, end_date=None, count=None) -> list:
         """获取交易日列表
 
@@ -269,13 +367,16 @@ class QuoteData:
         if dt_start or dt_end or count:
             if not dt_end:
                 dt_end = numpy.datetime64(datetime.date.today())
+            if dt_start:
+                result = result[numpy.where(result >= dt_start)]
+            if dt_end:
+                result = result[numpy.where(result <= dt_end)]
             if count and count > 0:
-                result = result[numpy.where(result <= dt_end)][-count:]
-            elif dt_start:
-                result = result[numpy.where((result >= dt_start) & (result <= dt_end))]
+                result = result[-count:]
         result = [datetime.date.fromtimestamp(x.astype(datetime.datetime) / 1e9) for x in result]
         return result
 
+    @assert_quote_init
     def get_security_info(self, security=None, fields=None, types=None) -> pandas.DataFrame:
         """获取证券信息数据
 
@@ -317,6 +418,7 @@ class QuoteData:
         result.set_index('code', inplace=True)
         return result
 
+    @assert_quote_init
     def get_security_xrxd(self, security, start_date=None, end_date=None, count=None) -> pandas.DataFrame:
         """获取证券除权除息数据
 
@@ -351,6 +453,7 @@ class QuoteData:
         df.reset_index(drop=True, inplace=True)
         return df
 
+    @assert_quote_init
     def get_quote_static(self, date=None, fields=None, exch=None) -> pandas.DataFrame:
         """获取静态行情数据
 
@@ -379,7 +482,6 @@ class QuoteData:
         fields = formatter_list(fields)
         exch = formatter_list(exch, ['SH', 'SZ'])
         curr_date = formatter_st(date, self.get_quote_date())
-        # print(curr_date)
         result = pandas.DataFrame()
         for x in exch:
             filename = os.path.join('static', x, time.strftime('%Y', curr_date),
@@ -389,30 +491,32 @@ class QuoteData:
                 continue
             df = df.copy()
             df['code'] = df['code'] + '.' + x
+            df['paused'] = df['paused'].astype('bool_')
             if fields:
                 cols_retain = ['datetime', 'code'] + fields
                 cols_all = list(df.columns)
                 cols_drop = [x for x in cols_all if x not in cols_retain]
                 df.drop(columns=cols_drop, inplace=True)
             result = result.append(df, ignore_index=True)
-        # print(result)
         return result
 
-    def get_quote_kdata(self, security, period='day', fq='pre', fq_vol=False, fields=None, start_date=None,
-                        end_date=None, count=None):
+    @assert_quote_init
+    def get_quote_kdata(self, security, period='day', fq='pre', fq_vol=False, fields=None,
+                        start_date=None, end_date=None, count=None, skip_paused=False):
         """获取K线数据
 
         参数:
             security: 一支股票代码或者一个股票代码的 list
-            period: 周期，默认值 day，支持 day, week, mon, season, year, mnX
-                    mnX 代表 X 分钟K线数据，如 mn1，mn5, mn15, mn30, mn60 等
-            fq: 复权方式；默认值 'pre'；'pre' 前复权, 'post' 后复权，None 不复权，复权日期
+            period: 周期，默认值 day，支持 day, daily, 1d, week, 1w, mon, month, 1M, season, year, mnX, Xm
+                    mnX 或 Xm 代表 X 分钟K线数据，如 mn1 mn5 mn15 mn30 mn60 或 1m 5m 15m 30m 60m 等
+            fq: 复权方式；默认值 'pre'；'pre' 前复权, 'post' 后复权，None 不复权，或 复权参考日
             fq_vol: 是否复权成交量：默认值 False
             fields: 获取的字段；类型 字符串 list；默认值 None 获取所有字段
                    可选值：'open','close','high','low','volume','money'
             start_date: 开始日期，与 count 二选一；类型 str/struct_time/datetime.date/datetime.datetime
             end_date: 结束日期；类型 str/struct_time/datetime.date/datetime.datetime；默认值 datetime.date.today()
             count: 结果集行数，与 start_date 二选一
+            skip_paused: 是否跳过停牌：默认值 False
 
         返回值:
             DataFrame 对象，包含 fields 指定的字段：
@@ -431,52 +535,18 @@ class QuoteData:
         security = formatter_list(security)
         fields = formatter_list(fields, ['open', 'close', 'high', 'low', 'volume', 'money'])
         start_date = formatter_st(start_date) if start_date else None
-        end_date = formatter_st(end_date, time.localtime())
+        end_date = formatter_st(end_date, self.get_quote_date())
         if not start_date and not count:
             raise ValueError('start_date or count must to be set.')
         elif start_date and count:
             raise ValueError('start_date and count cannot be set both.')
 
-        # 对周期进行分析，获得周期基类，day / mn1
-        if period in ['day', 'daily', 'week', 'mon', 'month', 'season', 'year']:
-            period_type = 'day'
-            period_step = 0
-        elif period[-1] == 'm' and period[:-1].isdecimal():
-            period_type = 'mn1'
-            period_step = int(period[:-1])
-        elif period[0:2] == 'mn' and period[2:].isdecimal():
-            period_type = 'mn1'
-            period_step = int(period[2:])
-        else:
-            raise Exception('period is invalid.')
-
+        # 对周期进行分析，获得周期类型与间隔，day / mn1
+        period_type, period_step = self._parse_period_str(period)
 
         # 针对设置 count，且周期基类为 day 的非单日周期，将 count 转换为 start_date 处理
-        if count and period_type == 'day' and period not in ['day', 'daily']:
-            if period == 'week':
-                trade_list = self.get_trade_days(count=count * 5)
-                week_maps = [x.isocalendar()[0] * 100 + x.isocalendar()[1] for x in trade_list]
-                week_list = list(set(week_maps))
-                week_list.sort()
-                w = week_list[-count]
-                start_date = trade_list[week_maps.index(w)].timetuple()
-            elif period in ['mon', 'month']:
-                y = end_date.tm_year
-                m = end_date.tm_mon + 1 - count
-                while m <= 0:
-                    y -= 1
-                    m += 12
-                start_date = datetime.datetime(y, m, 1).timetuple()
-            elif period == 'season':
-                y = end_date.tm_year
-                s = (end_date.tm_mon - 1) // 3 + 1 - count
-                while s < 0:
-                    y -= 1
-                    s += 4
-                start_date = datetime.datetime(y, s * 3 + 1, 1).timetuple()
-            elif period == 'year':
-                y = end_date.tm_year + 1 - count
-                start_date = datetime.datetime(y, 1, 1).timetuple()
+        if count and period_type == 'day' and period not in ['day', 'daily', '1d']:
+            start_date = self._parse_period_start(period, end_date, count)
             count = None
 
         # 需保留的列清单
@@ -502,13 +572,36 @@ class QuoteData:
         result = pandas.DataFrame()
         if mode_static:
             # 从静态文件获取数据
-            count = None
-            trade_list = self.get_trade_days(start_date, end_date, prep_count)
-            for x in trade_list:
-                df = self.get_quote_static(x)
-                df = df[df['code'].isin(security)]
+            check_security_count = count and skip_paused
+            if check_security_count:
+                sec_num = dict(zip(security, [0 for i in range(len(security))]))
+            sec_list = security
+            trade_list = self.get_trade_days(end_date=end_date)
+            for i in range(len(trade_list)):
+                dt = trade_list[-i - 1]
+                df = self.get_quote_static(dt)
+                # df = df.copy()
+                df = df[df['code'].isin(sec_list)]
+                if skip_paused:
+                    df = df[~df['paused']]
+                if check_security_count:
+                    for c in df['code'].tolist():
+                        if c not in sec_num:
+                            continue
+                        sec_num[c] += 1
+                        if sec_num[c] < count:
+                            continue
+                        sec_list.remove(c)
                 df = self.execute_rehabilitation(df, fq, fq_vol, cols_retain)
                 result = result.append(df, ignore_index=True)
+                if check_security_count:
+                    if not sec_list:
+                        break
+                else:
+                    if count and i + 1 >= count:
+                        break
+                if start_date and dt.timetuple() <= start_date:
+                    break
             cols_all = list(result.columns)
             cols_all.insert(0, cols_all.pop(cols_all.index('code')))
             result = result[cols_all]
@@ -519,18 +612,31 @@ class QuoteData:
             dt_start = numpy.datetime64(
                 datetime.datetime.fromtimestamp(time.mktime(start_date))) if start_date else None
             dt_end = numpy.datetime64(datetime.datetime.fromtimestamp(time.mktime(end_date)))
-            for x in security:
-                p = x.find('.')
+            append_paused = 'paused' in fields or skip_paused
+            for c in security:
+                p = c.find('.')
                 if p < 0:
                     continue
-                sec_code = x[0:p]
-                exc_code = x[p + 1:]
+                sec_code = c[0:p]
+                exc_code = c[p + 1:]
                 filename = os.path.join(period_type, exc_code, exc_code + '_' + period_type.upper()
                                         + '_' + sec_code + '.csv')
                 df = self._fetch_data(filename, date_columns='datetime')
                 if df is None:
                     continue
                 df = df.copy()
+                if append_paused:
+                    if period_type == 'day':
+                        df['paused'] = df.apply(lambda x: x['volume'] == 0 and x['money'] == 0, axis=1)
+                    elif period_type == 'mn1':
+                        c_period = 'day'
+                        c_file = os.path.join(c_period, exc_code, exc_code + '_' + c_period.upper()
+                                              + '_' + sec_code + '.csv')
+                        c_df = self._fetch_data(c_file, date_columns='datetime')
+                        c_skip_days = c_df[(c_df['volume'] == 0) & (c_df['money'] == 0)]['datetime'].tolist()
+                        df['paused'] = df['datetime'].astype('datetime64[D]').isin(c_skip_days)
+                if skip_paused:
+                    df = df[~df['paused']]
                 if not prep_count:
                     df = df[(df['datetime'] >= dt_start) & (df['datetime'] <= dt_end)]
                 else:
@@ -558,8 +664,7 @@ class QuoteData:
             agg_need = False
         if agg_need:
             agg_dict = {'datetime': 'last', 'open': 'first', 'close': 'last', 'high': 'max', 'low': 'min',
-                        'volume': 'sum',
-                        'money': 'sum'}
+                        'volume': 'sum', 'money': 'sum', 'paused': 'min'}
             for k in [x for x in agg_dict.keys() if x not in cols_retain]:
                 del agg_dict[k]
             result = result.groupby(['code', 'period']).agg(agg_dict)
@@ -576,3 +681,131 @@ class QuoteData:
         result.reset_index(drop=True, inplace=True)
 
         return result
+
+    def get_technical_analysis(self, security, indications, period='day', fq='pre', fq_vol=False, fields=None,
+                               start_date=None, end_date=None, count=None, skip_paused=False, keep_quote=False):
+        """获取技术指标
+
+        参数:
+            security: 一支股票代码或者一个股票代码的 list
+            indications: 技术指标及参与，单个或List
+                         单个指标：（'MA', 5)，
+                         多个指标：[('MA', 5), ('KDJ', 9, 3, 3)]
+                         可省略参数：['MA', 'KDJ']
+            period: 周期，默认值 day，支持 day, daily, 1d, week, 1w, mon, month, 1M, season, year, mnX, Xm
+                    mnX 或 Xm 代表 X 分钟K线数据，如 mn1 mn5 mn15 mn30 mn60 或 1m 5m 15m 30m 60m 等
+            fq: 复权方式；默认值 'pre'；'pre' 前复权, 'post' 后复权，None 不复权，或 复权参考日
+            fq_vol: 是否复权成交量：默认值 False
+            fields: 获取的字段；类型 字符串 list；默认值 None 获取所有字段
+                   可选值：'open','close','high','low','volume','money'
+            start_date: 开始日期，与 count 二选一；类型 str/struct_time/datetime.date/datetime.datetime
+            end_date: 结束日期；类型 str/struct_time/datetime.date/datetime.datetime；默认值 datetime.date.today()
+            count: 结果集行数，与 start_date 二选一
+            skip_paused: 是否跳过停牌：默认值 False
+            keep_quote: 是否保留行情字段；默认值 False
+
+        返回值:
+            DataFrame 对象，字段为指标名称
+
+        """
+
+        # 计算时需要预加载的数量
+        refer_num = 500
+        start_date = formatter_st(start_date) if start_date else None
+        end_date = formatter_st(end_date, self.get_quote_date())
+        if not start_date and not count:
+            raise ValueError('start_date or count must to be set.')
+        elif start_date and count:
+            raise ValueError('start_date and count cannot be set both.')
+        prep_start = None
+        prep_count = None
+        if start_date:
+            prep_start = self._parse_period_start(period, start_date, refer_num)
+        elif count:
+            prep_count = count + refer_num
+
+        indications = formatter_list(indications)
+        cols_quote = ['open', 'close', 'high', 'low', 'volume', 'money', 'paused']
+        fields = formatter_list(fields, cols_quote)
+        df = self.get_quote_kdata(security, period=period, fq=fq, fq_vol=True, fields=cols_quote
+                                  , start_date=prep_start, end_date=end_date, count=prep_count)
+        for v in indications:
+            v = [v] if type(v) is str else list(v)
+            c = v.pop(0)
+            func = globals().get(c)
+            if not func:
+                print('%s is not exists.' % c)
+                continue
+            if not hasattr(func, '__wrapped__'):
+                print('%s is unavailable.' % c)
+                continue
+            f_code = getattr(func, '__wrapped__').__code__
+            f_args = [x for x in f_code.co_varnames[:f_code.co_argcount]]
+            if len(f_args) > 0 and f_args[0] == 'X' and len(v) > 0 and type(v[0]) is not str:
+                f_args.pop(0)
+            kw = dict(axis=1)
+            for i in range(len(v)):
+                if i >= len(f_args):
+                    break
+                kw[f_args[i]] = v[i]
+
+            #获取返回值字段名
+            cols_fn = arithmetic_returns(f_code)
+            if not cols_fn:
+                cols_fn = c
+            if type(cols_fn) is str:
+                cols_fn = [cols_fn]
+            if not cols_fn:
+                raise Exception('columns name not support.')
+
+            # 字段重名处理
+            if [x for x in cols_fn if x in df.columns]:
+                c_arr = [str(x) for x in v]
+                if len(cols_fn) > 1:
+                    c_arr.insert(0, c)
+                    cols_fn = ['_'.join(c_arr) + '_' + x for x in cols_fn]
+                elif len(cols_fn) == 1:
+                    if cols_fn[0] == c:
+                        c_arr.insert(0, c)
+                        cols_fn = ['_'.join(c_arr)]
+                    else:
+                        c_arr.insert(0, cols_fn[0])
+                        cols_fn = ['_'.join(c_arr)]
+
+            # 提交指标函数处理
+            arithmetic_reset()
+            se = df.apply(func, **kw)
+            df[cols_fn] = pandas.DataFrame(se.tolist(), index=df.index)
+
+        # 跳过停牌
+        if skip_paused:
+            df = df[~df['paused']]
+
+        # 结果集数量显示
+        result = pandas.DataFrame()
+        if not count:
+            dt_start = numpy.datetime64(
+                datetime.datetime.fromtimestamp(time.mktime(start_date))) if start_date else None
+            dt_end = numpy.datetime64(datetime.datetime.fromtimestamp(time.mktime(end_date)))
+            result = result.append(df[(df['datetime'] >= dt_start) & (df['datetime'] <= dt_end)], ignore_index=True)
+        else:
+            c_groupby = df.groupby(by='code')
+            for x, y in c_groupby:
+                result = result.append(y.iloc[-count:], ignore_index=True)
+        result.reset_index(drop=True, inplace=True)
+
+        # 是否保留行情字段
+        cols_drop = []
+        if not keep_quote:
+            cols_drop = cols_quote
+        elif type(keep_quote) is str:
+            cols_drop = [x for x in cols_quote if x != keep_quote]
+        elif type(keep_quote) is list:
+            cols_drop = [x for x in cols_quote if x not in keep_quote]
+        elif keep_quote:
+            cols_drop = [x for x in cols_quote if x not in fields]
+        cols_drop = list(set(cols_drop).intersection(set(result.columns)))
+        if cols_drop:
+            result.drop(columns=cols_drop, inplace=True)
+
+        return result
